#include <fmt/core.h>
#include <fmt/ranges.h>
#include <stdio.h>

#include <CLI/CLI.hpp>
#include <optional>

#include "utils.h"

using json = nlohmann::json;
using std::nullopt;
using std::optional;
using std::string;
using std::vector;

auto tables = ExistTables("./tables.json");

void create_table(const json& j) {
    auto table_name = table_name_of(j);

    if (tables.exists(table_name)) {
        throw std::runtime_error(
            fmt::format("table `{}` exists\n", table_name));
    }
    tables.set(table_name, j["cols"]);
    tables.save();
}

void check_where_clause(const vector<string> table_names, const json& j) {
    if (j.empty()) {
        return;
    }

    auto type = type_of(j);
    auto left = j["left"];
    auto right = j["right"];
    for (const auto curr_branch_name : {"left", "right"}) {
        auto curr_branch = j[curr_branch_name];
        if (is_column(curr_branch)) {
            // 说明到叶节点了
            auto col_def =
                tables.get_select_column_define(table_names, curr_branch);
            if (!col_def.has_value()) {
                throw std::runtime_error(
                    fmt::format("column `{}` not exists in `{}`\n",
                                select_column_get_name(curr_branch),
                                fmt::join(table_names, ", ")));
            } else {
                if (curr_branch_name == "right" &&
                    op_type_of(j) == OpType::OP_COMPARE) {
                    // 如果左右分支都是普通列, 并且当前节点操作符是比较,
                    // 那么应当判断类型是否一致 如果能走到right,
                    // 说明left也是普通列
                    let curr_left_col_def =
                        tables.get_select_column_define(table_names, left);
                    if (curr_left_col_def.value().data_type !=
                        col_def.value().data_type) {
                        throw std::runtime_error(
                            fmt::format("column `{}` type is `{}`, but `{}` "
                                        "type is `{}`, cannot `{}` \n",
                                        select_column_get_name(left),
                                        curr_left_col_def.value().data_type,
                                        select_column_get_name(curr_branch),
                                        col_def.value().data_type, j["type"]));
                    }
                }
            }

        } else {
            check_where_clause(table_names, curr_branch);
        }
    }
}

void check_select_table(const vector<string> table_names,
                        const json& select_cols) {
    for (let& select_col : select_cols) {
        let col_def =
            tables.get_select_column_define(table_names, select_col["target"]);

        if (!col_def.has_value()) {
            throw std::runtime_error(
                fmt::format("column `{}` not exists in `{}`\n",
                            select_column_get_name(select_col["target"]),
                            fmt::join(table_names, ", ")));
        }
    }
}

vector<string> check_join_table(const vector<string> table_names,
                                const json& join_options) {
    let join_table_name = join_options["join_with"]["value"];
    auto new_table_names = vector(table_names);
    new_table_names.push_back(join_table_name);

    if (!tables.exists(join_table_name)) {
        throw std::runtime_error(
            fmt::format("table `{}` not exists\n", join_table_name));
    }

    check_where_clause(new_table_names, join_options["on"]);
    return new_table_names;
}

void process_sql(json& j) {
    let type = j.value("type", "none");

    if (type == "create_stmt") {
        // 创建表
        create_table(j);
    } else if (type == "drop_stmt") {
        if (tables.exists(j["table_name"])) {
            tables.remove(j["table_name"]);
            tables.save();
        }
        return;
    } else if (type == "select_stmt") {
        vector<string> table_names = table_names_of(j);
        for (let table_name : table_names) {
            if (!tables.exists(table_name)) {
                fmt::print("error: table `{}` does not exist\n", table_name);
                return;
            }
        }
        if (j.contains("join_options")) {
            auto new_tables = check_join_table(table_names, j["join_options"]);
            table_names = new_tables;
        }

        check_select_table(table_names, j["select_cols"]);

        check_where_clause(table_names, j["where"]);
    } else {
        throw std::runtime_error(fmt::format(
            "Unknown expression type `{}` total: \n{}", type, j.dump(2)));
    }
}

int main(int argc, char** argv) {
    CLI::App app{"sql 检查器, 附带创建表相关功能"};
    string sql = "", filename = "";

    app.add_option("-f,--file", filename, "输入的 sql 解析后的 json 文件")
        ->check(CLI::ExistingFile);
    app.add_option("-s,--sql", sql, "输入的 sql");

    CLI11_PARSE(app, argc, argv);

    json stmts;
    if (sql.length() > 0) {
        stmts = parse_sql(sql).out;
    } else if (filename.length() > 0) {
        stmts = json::parse(std::ifstream(filename));
    } else {
        std::cout << "请输入 sql 或者 sql 文件" << std::endl;
        return 1;
    }

    for (auto& stmt : stmts) {
        // for(auto& item : stmt.items()) {
        //     fmt::print("{}: {}\n", item.key(), item.value().dump());
        // }
        process_sql(stmt);
    }

    return 0;
}
